/*
 * JCuda - Java bindings for NVIDIA CUDA driver and runtime API
 * http://www.jcuda.org
 *
 * Copyright 2009-2012 Marco Hutter - http://www.jcuda.org
 */

package org.jcuda.samples.matrix.cg;

/**
 * Default implementation of a {@link FloatMatrixHostCSR}
 */
class DefaultFloatMatrixHostCSR implements FloatMatrixHostCSR
{
    /**
     * The number of rows 
     */
    private final int numRows;
    
    /**
     * The number of columns 
     */
    private final int numCols;

    /**
     * The number of non-zero elements in this matrix
     */
    private final int numNonZeros;
    
    /**
     * The non-zero values 
     */
    private final float values[];
    
    /**
     * The row pointers 
     */
    private final int rowPointers[];
    
    /**
     * The column indices 
     */
    private final int columnIndices[];
    
    /**
     * Creates a new matrix with the given values. 
     * 
     * @param numRows The number of rows 
     * @param numCols The number of columns
     * @param numNonZeros The number of non-zero elements 
     * @param values The non-zero values
     * @param rowPointers The row pointers
     * @param columnIndices The column indices
     */
    DefaultFloatMatrixHostCSR(int numRows, int numCols, int numNonZeros)
    {
        this.numRows = numRows;
        this.numCols = numCols;
        this.numNonZeros = numNonZeros;
        this.values = new float[numNonZeros];
        this.rowPointers = new int[numRows+1];
        this.columnIndices = new int[numNonZeros];
    }
    
    /**
     * Creates a new matrix with the given values. All the given
     * values will be stored directly as a reference. 
     * 
     * @param numRows The number of rows 
     * @param numCols The number of columns
     * @param numNonZeros The number of non-zero elements 
     * @param values The non-zero values
     * @param rowPointers The row pointers
     * @param columnIndices The column indices
     */
    DefaultFloatMatrixHostCSR(
        int numRows, int numCols, int numNonZeros, 
        float[] values, int[] rowPointers, int[] columnIndices)
    {
        this.numRows = numRows;
        this.numCols = numCols;
        this.numNonZeros = numNonZeros;
        this.values = values;
        this.rowPointers = rowPointers;
        this.columnIndices = columnIndices;
    }
    
    @Override
    public int getNumNonZeros()
    {
        return numNonZeros;
    }

    @Override
    public float[] getValues()
    {
        return values;
    }

    @Override
    public int[] getRowPointers()
    {
        return rowPointers;
    }

    @Override
    public int[] getColumnIndices()
    {
        return columnIndices;
    }
    
    @Override
    public int getNumRows()
    {
        return numRows;
    }

    @Override
    public int getNumCols()
    {
        return numCols;
    }

    @Override
    public float get(int r, int c)
    {
        int c0 = rowPointers[r];
        int c1 = rowPointers[r+1];
        for (int cc=c0; cc<c1; cc++)
        {
            if (columnIndices[cc] == c)
            {
                return values[cc];
            }
        }
        return 0;
    }
    
    
    @Override
    public String toString()
    {
        return Matrices.toString2D(this);
    }
}