/*
 * JCuda - Java bindings for NVIDIA CUDA driver and runtime API
 * http://www.jcuda.org
 *
 * Copyright 2009-2012 Marco Hutter - http://www.jcuda.org
 */

package org.jcuda.samples.matrix.cg;

import static jcuda.jcusparse.JCusparse.*;
import static jcuda.jcusparse.cusparseDiagType.CUSPARSE_DIAG_TYPE_NON_UNIT;
import static jcuda.jcusparse.cusparseFillMode.CUSPARSE_FILL_MODE_LOWER;
import static jcuda.jcusparse.cusparseIndexBase.CUSPARSE_INDEX_BASE_ZERO;
import static jcuda.jcusparse.cusparseMatrixType.*;

import java.util.Locale;

import jcuda.driver.CUdeviceptr;
import jcuda.jcusparse.*;

/**
 * Utility methods for matrix creation and conversion
 */
public class Matrices
{
    /**
     * Create a dense float matrix that is stored in host memory
     * 
     * @param numRows The number of rows
     * @param numCols The number of columns
     * @return The new matrix
     */
    public static FloatMatrixHostDense createHostDense(
        int numRows, int numCols)
    {
        DefaultFloatMatrixHostDense result = 
            new DefaultFloatMatrixHostDense(numRows, numCols);
        return result;
    }
    
    /**
     * Create a dense float matrix that is stored in host memory and
     * has the same contents as the given matrix.
     * 
     * @param matrix The input matrix
     * @return The new matrix
     */
    public static FloatMatrixHostDense createHostDense(FloatMatrixHost matrix)
    {
        int numRows = matrix.getNumRows();
        int numCols = matrix.getNumCols();
        DefaultFloatMatrixHostDense result = 
            new DefaultFloatMatrixHostDense(numRows, numCols);
        for (int r=0; r<numRows; r++)
        {
            for (int c=0; c<numCols; c++)
            {
                float value = matrix.get(r, c);
                result.set(r, c, value);
            }
        }
        return result;
    }

    
    /**
     * Create a float matrix in CSR format that is stored in host memory
     * and has the same contents as the given matrix.
     * Only elements whose absolute value is greater than or equal
     * to the given epsilon will be considered as being 'non-zero'. 
     * 
     * @param matrix The input matrix
     * @param epsilon The epsilon for non-zero values
     * @return The new matrix
     */
    public static FloatMatrixHostCSR createHostCSR(
        FloatMatrixHost matrix, float epsilon)
    {
        int numNonZeros = countNumNonZeros(matrix, epsilon);
        int numRows = matrix.getNumRows();
        int numCols = matrix.getNumCols();
        float values[] = new float[numNonZeros];
        int columnIndices[] = new int[numNonZeros];
        int rowPointers[] = new int[numRows+1];
        rowPointers[rowPointers.length-1] = numNonZeros;
        
        int index = 0;
        for (int r=0; r<numRows; r++)
        {
            boolean firstColumn = true;
            for (int c=0; c<numCols; c++)
            {
                float value = matrix.get(r, c);
                if (Math.abs(value) >= epsilon)
                {
                    if (firstColumn)
                    {
                        rowPointers[r] = index;
                        firstColumn = false;
                    }
                    values[index] = value;
                    columnIndices[index] = c;
                    index++;
                }
            }
        }
        return new DefaultFloatMatrixHostCSR(
            numRows, numCols, numNonZeros, 
            values, rowPointers, columnIndices);
        
    }
    
    /**
     * Count the number of non-zero elements in the given matrix.
     * Only elements whose absolute value is greater than or equal
     * to the given epsilon will be counted. 
     * 
     * @param matrix The matrix
     * @param epsilon The epsilon
     * @return The number of non-zero elements
     */
    static int countNumNonZeros(FloatMatrixHost matrix, float epsilon)
    {
        int count = 0;
        for (int r=0; r<matrix.getNumRows(); r++)
        {
            for (int c=0; c<matrix.getNumCols(); c++)
            {
                float value = matrix.get(r, c);
                if (Math.abs(value) >= epsilon)
                {
                    count++;
                }
            }
        }
        return count;
    }
    
    
    
    /**
     * Create a dense float matrix that is stored in host memory and
     * has the same contents as the given dense device memory matrix
     * 
     * @param matrix The input matrix
     * @return The new matrix
     */
    public static FloatMatrixHostDense createHostDense(
        FloatMatrixDeviceDense matrix)
    {
        int numRows = matrix.getNumRows();
        int numCols = matrix.getNumCols();
        float[] data = 
            Utils.createFloatArray(matrix.getData(), numRows * numCols);
        return new DefaultFloatMatrixHostDense(numRows, numCols, data);
    }
    
    /**
     * Create a float matrix in CSR format that is stored in host memory
     * and has the same contents as the given CSR device memory matrix.
     * 
     * @param matrix The input matrix
     * @return The new matrix
     */
    public static FloatMatrixHostCSR createHostCSR(
        FloatMatrixDeviceCSR matrix)
    {
        int numRows = matrix.getNumRows();
        int numCols = matrix.getNumCols();
        int numNonZeros = matrix.getNumNonZeros();
        float[] values = 
            Utils.createFloatArray(matrix.getValues(), numNonZeros);
        int[] rowPointers = 
            Utils.createIntArray(matrix.getRowPointers(), numRows + 1);
        int[] columnIndices = 
            Utils.createIntArray(matrix.getColumnIndices(), numNonZeros);
        return new DefaultFloatMatrixHostCSR(
            numRows, numCols, numNonZeros, 
            values, rowPointers, columnIndices);
    }
    
    
    /**
     * Create a float matrix in CSR format that is stored in device memory
     * and has the same contents as the given matrix that is stored in 
     * host memory.
     * Only elements whose absolute value is greater than or equal
     * to the given epsilon will be considered as being 'non-zero'. 
     * 
     * @param handle The CUSPARSE context handle
     * @param matrix The input matrix.
     * @param epsilon The epsilon for non-zero values
     * @return The new matrix
     */
    public static FloatMatrixDeviceCSR createDeviceCSR(
        cusparseHandle handle, FloatMatrixHostDense matrix, float epsilon)
    {
        return createDeviceCSR(handle, createHostCSR(matrix, epsilon));
    }
    
    /**
     * Create a float matrix in CSR format that is stored in device memory
     * and has the same contents as the given matrix that is stored in 
     * host memory.
     * 
     * @param handle The CUSPARSE context handle
     * @param matrix The input matrix.
     * @return The new matrix
     */
    public static FloatMatrixDeviceCSR createDeviceCSR(
        cusparseHandle handle, FloatMatrixHostCSR matrix)
    {
        cusparseMatDescr descr = new cusparseMatDescr();
        cusparseCreateMatDescr(descr);
        cusparseSetMatType(descr, CUSPARSE_MATRIX_TYPE_GENERAL);
        cusparseSetMatIndexBase(descr, CUSPARSE_INDEX_BASE_ZERO);

        int numRows = matrix.getNumRows();
        int numCols = matrix.getNumCols();
        int numNonZeros = matrix.getNumNonZeros();
        
        CUdeviceptr values = 
            Utils.createPointer(matrix.getValues());
        CUdeviceptr rowPointers = 
            Utils.createPointer(matrix.getRowPointers());
        CUdeviceptr columnIndices = 
            Utils.createPointer(matrix.getColumnIndices());

        return new DefaultFloatMatrixDeviceCSR(
            descr, numRows, numCols, numNonZeros, 
            values, rowPointers, columnIndices);
    }
    
    
    
    /**
     * Generate the incomplete Cholesky factor H (lower triangular) for 
     * the given matrix, in CSR format, stored in device memory. <br />
     * <br />
     * <strong>NOTE:</strong> This method is rather inefficient, because 
     * it copies the data from the device to the host, creates the result 
     * matrix in host memory, and copies the data back from the host to 
     * the device. This could be optimized by using an own kernel for
     * the creation of the ICP matrix.
     *  
     * @param cusparseHandle The CUSPARSE context handle
     * @param matrix The input matrix
     * @return The new matrix
     */
    static FloatMatrixDeviceCSR createDeviceICP(
        cusparseHandle cusparseHandle, FloatMatrixDeviceCSR matrix)
    {
        FloatMatrixHostCSR matrixHostCSR = Matrices.createHostCSR(matrix);
        int row[] = matrixHostCSR.getRowPointers();
        float val[] = matrixHostCSR.getValues();
        
        int numRows = matrix.getNumRows();
        int numNonZerosICP = 2 * numRows - 1;
        FloatMatrixHostCSR hostICP = 
            new DefaultFloatMatrixHostCSR(numRows, numRows, numNonZerosICP);
        genICP(row, val, numRows, hostICP.getColumnIndices(), 
            hostICP.getRowPointers(), hostICP.getValues());
        
        cusparseMatDescr descr = new cusparseMatDescr();
        cusparseCreateMatDescr(descr);
        cusparseSetMatType(descr, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
        cusparseSetMatIndexBase(descr, CUSPARSE_INDEX_BASE_ZERO);
        cusparseSetMatDiagType(descr, CUSPARSE_DIAG_TYPE_NON_UNIT);
        cusparseSetMatFillMode(descr, CUSPARSE_FILL_MODE_LOWER);
        
        CUdeviceptr values = 
            Utils.createPointer(hostICP.getValues());
        CUdeviceptr rowPointers = 
            Utils.createPointer(hostICP.getRowPointers());
        CUdeviceptr columnIndices = 
            Utils.createPointer(hostICP.getColumnIndices());
        
        return new DefaultFloatMatrixDeviceCSR(
            descr, numRows, numRows, numNonZerosICP, 
            values, rowPointers, columnIndices);
    }
    
    /**
     * Generate the Incomplete Cholesky factor H (lower triangular).
     * 
     * Adapted from the NVIDIA conjugate gradient sample.
     * 
     * @param rowPointers The row pointers
     * @param values The values
     * @param numRows The number of rows
     * @param columnIndicesICP The column indices of the ICP
     * @param rowPointersICP The row pointers of the ICP
     * @param valuesICP The value of the ICP
     */
    private static void genICP(int rowPointers[], float values[], int numRows, 
        int columnIndicesICP[], int rowPointersICP[], float valuesICP[])
    {
        rowPointersICP[0] = 0;
        columnIndicesICP[0] = 0;
        int inz = 1;

        for (int k = 1; k < numRows; k++)
        {
            rowPointersICP[k] = inz;
            for (int j = k - 1; j <= k; j++)
            {
                columnIndicesICP[inz] = j;
                inz++;
            }
        }
        rowPointersICP[numRows] = inz;

        valuesICP[0] = values[0];
        for (int k = 1; k < numRows; k++)
        {
            int rp = rowPointers[k];
            int rpICP = rowPointersICP[k];
            valuesICP[rpICP] = values[rp];
            valuesICP[rpICP + 1] = values[rp + 1];
        }

        for (int k = 0; k < numRows; k++)
        {
            int rpICP = rowPointersICP[k + 1];
            valuesICP[rpICP - 1] = (float)Math.sqrt(valuesICP[rpICP - 1]);
            if (k < numRows - 1)
            {
                valuesICP[rpICP] /= valuesICP[rpICP - 1];
                valuesICP[rpICP + 1] -= valuesICP[rpICP] * valuesICP[rpICP];
            }
        }
    }
    

    
    /**
     * Creates a String representation of the given matrix. If the
     * matrix is too large, only the upper left part will be printed.    
     *
     * @param matrix The matrix
     * @return The String representation
     */
    public static String toString2D(FloatMatrixHost matrix)
    {
        return toString2D(matrix, 16);
    }

    /**
     * Creates a String representation of the given matrix. If the
     * matrix is too large, only the upper left part will be printed.    
     *
     * @param matrix The matrix
     * @param maxSize The maximum size of the printed matrix
     * @return The String representation
     */
    public static String toString2D(FloatMatrixHost matrix, int maxSize)
    {
        String format = "%6.3f  ";
        StringBuilder sb = new StringBuilder();
        for (int r=0; r<matrix.getNumRows(); r++)
        {
            for (int c=0; c<matrix.getNumCols(); c++)
            {
                sb.append(String.format(Locale.ENGLISH, 
                    format, matrix.get(r,c)));
                
                if (c == maxSize)
                {
                    sb.append("...");
                    break;
                }
            }
            sb.append("\n");
            if (r == maxSize)
            {
                sb.append("...");
                break;
            }
        }
        return sb.toString();
    }
    
    /**
     * Private constructor to prevent instantiation
     */
    private Matrices()
    {
        
    }
    
}
