/*
 * JCuda - Java bindings for NVIDIA CUDA driver and runtime API
 * http://www.jcuda.org
 *
 * Copyright 2009-2012 Marco Hutter - http://www.jcuda.org
 */

package org.jcuda.samples.matrix.cg;

import static jcuda.runtime.JCuda.*;
import static jcuda.runtime.cudaMemcpyKind.*;
import jcuda.*;
import jcuda.driver.CUdeviceptr;

/**
 * Utility methods for memory transfers
 */
public class Utils
{
    /**
     * Create an array from the given device data with the given size
     *
     * @param deviceData The device data
     * @param size The size (in number of elements)
     * @return The new array
     */
    public static float[] createFloatArray(Pointer deviceData, int size)
    {
        float hostData[] = new float[size];
        cudaMemcpy(Pointer.to(hostData), deviceData,
            size * Sizeof.FLOAT, cudaMemcpyDeviceToHost);
        return hostData;
    }

    /**
     * Create an array from the given device data with the given size
     *
     * @param deviceData The device data
     * @param size The size (in number of elements)
     * @return The new array
     */
    public static int[] createIntArray(Pointer deviceData, int size)
    {
        int hostData[] = new int[size];
        cudaMemcpy(Pointer.to(hostData), deviceData,
            size * Sizeof.INT, cudaMemcpyDeviceToHost);
        return hostData;
    }

    /**
     * Create a pointer to device memory that has the same contents
     * as the given array.
     *
     * @param hostData The host data
     * @return The new pointer
     */
    public static CUdeviceptr createPointer(float hostData[])
    {
        CUdeviceptr deviceData = new CUdeviceptr();
        cudaMalloc(deviceData, hostData.length * Sizeof.FLOAT);
        cudaMemcpy(deviceData, Pointer.to(hostData),
            hostData.length * Sizeof.FLOAT, cudaMemcpyHostToDevice);
        return deviceData;
    }

    /**
     * Create a pointer to device memory that has the same contents
     * as the given array.
     *
     * @param hostData The host data
     * @return The new pointer
     */
    public static CUdeviceptr createPointer(int hostData[])
    {
        CUdeviceptr deviceData = new CUdeviceptr();
        cudaMalloc(deviceData, hostData.length * Sizeof.INT);
        cudaMemcpy(deviceData, Pointer.to(hostData),
            hostData.length * Sizeof.INT, cudaMemcpyHostToDevice);
        return deviceData;
    }

    /**
     * Copy the contents of the given array to the given device pointer
     *
     * @param hostData The host data
     * @param deviceData The device data
     */
    public static void copyToDevice(float hostData[], Pointer deviceData)
    {
        cudaMemcpy(deviceData, Pointer.to(hostData),
            hostData.length * Sizeof.FLOAT, cudaMemcpyHostToDevice);
    }

    /**
     * Copy the contents of the given device pointer to the given array
     *
     * @param deviceData The device data
     * @param hostData The host data
     */
    public static void copyToHost(Pointer deviceData, float hostData[])
    {
        cudaMemcpy(Pointer.to(hostData), deviceData,
            hostData.length * Sizeof.FLOAT, cudaMemcpyDeviceToHost);
    }



    /**
     * Private constructor to prevent instantiation
     */
    private Utils()
    {

    }
}
